@@ -41,6 +41,8 @@ int Detail(char *dev, int brief, int test)
 	int fd = open(dev, O_RDONLY, 0);
 	int vers;
 	mdu_array_info_t array;
+	mdu_disk_info_t *disks;
+	int next;
 	int d;
 	time_t atime;
 	char *c;
@@ -50,9 +52,10 @@ int Detail(char *dev, int brief, int test)
 	int is_26 = get_linux_version() >= 2006000;
 	int is_rebuilding = 0;
 	int failed = 0;
+	struct supertype *st = NULL;
+	int max_disks = MD_SB_DISKS;
 
-	mdp_super_t super;
-	int have_super = 0;
+	void *super = NULL;
 	int rv = test ? 4 : 1;
 
 	if (fd < 0) {
@@ -83,12 +86,16 @@ int Detail(char *dev, int brief, int test)
 		close(fd);
 		return rv;
 	}
+	st = super_by_version(array.major_version, array.minor_version);
+
 	if (fstat(fd, &stb) != 0 && !S_ISBLK(stb.st_mode))
 		stb.st_rdev = 0;
 	rv = 0;
 
+	if (st) max_disks = st->max_devs;
+
 	/* try to load a superblock */
-	for (d= 0; d<MD_SB_DISKS; d++) {
+	for (d= 0; d<max_disks; d++) {
 		mdu_disk_info_t disk;
 		char *dv;
 		disk.number = d;
@@ -99,16 +106,22 @@ int Detail(char *dev, int brief, int test)
 		    disk.minor == 0)
 			continue;
 		if ((dv=map_dev(disk.major, disk.minor))) {
-			if (!have_super && (disk.state & (1<<MD_DISK_ACTIVE))) {
+			if (!super && (disk.state & (1<<MD_DISK_ACTIVE))) {
 				/* try to read the superblock from this device
 				 * to get more info
 				 */
 				int fd2 = open(dv, O_RDONLY);
-				if (fd2 >=0 &&
-				    load_super(fd2, &super) ==0 &&
-				    (unsigned long)super.ctime == (unsigned long)array.ctime &&
-				    (unsigned int)super.level == (unsigned int)array.level)
-					have_super = 1;
+				if (fd2 >=0 && st &&
+				    st->ss->load_super(st, fd2, &super, NULL) == 0) {
+					struct mdinfo info;
+					struct mddev_ident_s ident;
+					st->ss->getinfo_super(&info, &ident, super);
+					if (info.array.ctime != array.ctime ||
+					    info.array.level != array.level) {
+						free(super);
+						super = NULL;
+					}
+				}
 				if (fd2 >= 0) close(fd2);
 			}
 		}
@@ -119,9 +132,10 @@ int Detail(char *dev, int brief, int test)
 	if (brief) 
 		printf("ARRAY %s level=%s num-devices=%d", dev, c?c:"-unknown-",array.raid_disks );
 	else {
+		mdu_bitmap_file_t bmf;
 		unsigned long array_size;
 		unsigned long long larray_size;
-		struct mdstat_ent *ms = mdstat_read(0);
+		struct mdstat_ent *ms = mdstat_read(0, 0);
 		struct mdstat_ent *e;
 		int devnum = array.md_minor;
 		if (major(stb.st_rdev) != MD_MAJOR)
@@ -150,14 +164,29 @@ int Detail(char *dev, int brief, int test)
 		printf("     Raid Level : %s\n", c?c:"-unknown-");
 		if (larray_size)
 			printf("     Array Size : %llu%s\n", (larray_size>>10), human_size(larray_size));
-		if (array.level >= 1)
-			printf("    Device Size : %d%s\n", array.size, human_size((long long)array.size<<10));
+		if (array.level >= 1) {
+			if (array.major_version != 0 &&
+			    (larray_size >= 0xFFFFFFFFULL|| array.size == 0)) {
+				unsigned long long dsize = get_component_size(fd);
+				if (dsize > 0)
+					printf("    Device Size : %llu%s\n", dsize, human_size((long long)array.size<<10));
+				else
+					printf("    Device Size : unknown\n");
+			} else
+				printf("    Device Size : %d%s\n", array.size, human_size((long long)array.size<<10));
+		}
 		printf("   Raid Devices : %d\n", array.raid_disks);
 		printf("  Total Devices : %d\n", array.nr_disks);
 		printf("Preferred Minor : %d\n", array.md_minor);
 		printf("    Persistence : Superblock is %spersistent\n",
 		       array.not_persistent?"not ":"");
 		printf("\n");
+		if (ioctl(fd, GET_BITMAP_FILE, &bmf) == 0 &&
+			bmf.pathname[0]) {
+			printf("  Intent Bitmap : %s\n", bmf.pathname);
+			printf("\n");
+		} else if (array.state & (1<<MD_SB_BITMAP_PRESENT))
+			printf("  Intent Bitmap : Internal\n\n");
 		atime = array.utime;
 		printf("    Update Time : %.24s\n", ctime(&atime));
 		printf("          State : %s%s%s\n",
@@ -198,21 +227,20 @@ int Detail(char *dev, int brief, int test)
 		}
 		free_mdstat(ms);
 
-		if (have_super) {
-			printf("           UUID : ");
-			if (super.minor_version >= 90)
-				printf("%08x:%08x:%08x:%08x", super.set_uuid0, super.set_uuid1,
-				       super.set_uuid2, super.set_uuid3);
-			else
-				printf("%08x", super.set_uuid0);
-			printf("\n         Events : %d.%d\n\n", super.events_hi, super.events_lo);
-		}
+		if (super && st)
+			st->ss->detail_super(super);
 
 		printf("    Number   Major   Minor   RaidDevice State\n");
 	}
-	for (d= 0; d<MD_SB_DISKS; d++) {
+	disks = malloc(max_disks * sizeof(mdu_disk_info_t));
+	for (d=0; d<max_disks; d++) {
+		disks[d].state = (1<<MD_DISK_REMOVED);
+		disks[d].major = disks[d].minor = 0;
+	}
+
+	next = array.raid_disks;
+	for (d=0; d < max_disks; d++) {
 		mdu_disk_info_t disk;
-		char *dv;
 		disk.number = d;
 		if (ioctl(fd, GET_DISK_INFO, &disk) < 0) {
 			if (d < array.raid_disks)
@@ -220,12 +248,24 @@ int Detail(char *dev, int brief, int test)
 					d, strerror(errno));
 			continue;
 		}
+		if (disk.major == 0 && disk.minor == 0)
+			continue;
+		if (disk.raid_disk >= 0 && disk.raid_disk < array.raid_disks) 
+			disks[disk.raid_disk] = disk;
+		else if (next < max_disks)
+			disks[next++] = disk;
+	}
+
+	for (d= 0; d < max_disks; d++) {
+		char *dv;
+		mdu_disk_info_t disk = disks[d];
+
 		if (d >= array.raid_disks &&
 		    disk.major == 0 &&
 		    disk.minor == 0)
 			continue;
 		if (!brief) {
-			if (disk.number == array.raid_disks) printf("\n");
+			if (d == array.raid_disks) printf("\n");
 			if (disk.raid_disk < 0)
 				printf("   %5d   %5d    %5d        -     ", 
 				       disk.number, disk.major, disk.minor);
@@ -241,8 +281,11 @@ int Detail(char *dev, int brief, int test)
 			if (disk.state & (1<<MD_DISK_ACTIVE)) printf(" active");
 			if (disk.state & (1<<MD_DISK_SYNC)) printf(" sync");
 			if (disk.state & (1<<MD_DISK_REMOVED)) printf(" removed");
-			if (disk.state == 0) printf(" spare");
-			if (disk.state == 0) {
+			if (disk.state & (1<<MD_DISK_WRITEMOSTLY)) printf(" writemostly");
+			if ((disk.state &
+			     ((1<<MD_DISK_ACTIVE)|(1<<MD_DISK_SYNC)|(1<<MD_DISK_REMOVED)))
+			    == 0) {
+				printf(" spare");
 				if (is_26) {
 					if (disk.raid_disk < array.raid_disks && disk.raid_disk >= 0)
 						printf(" rebuilding");
@@ -278,14 +321,9 @@ int Detail(char *dev, int brief, int test)
 		if (!brief) printf("\n");
 	}
 	if (spares && brief) printf(" spares=%d", spares);
-	if (have_super && brief) {
-		printf(" UUID=");
-	    	if (super.minor_version >= 90)
-			printf("%08x:%08x:%08x:%08x", super.set_uuid0, super.set_uuid1,
-			       super.set_uuid2, super.set_uuid3);
-		else
-			printf("%08x", super.set_uuid0);
-	}
+	if (super && brief && st)
+		st->ss->brief_detail_super(super);
+
 	if (brief > 1 && devices) printf("\n   devices=%s", devices);
 	if (brief) printf("\n");
 	if (test && (rv&2)) rv &= ~1;
