@@ -2,7 +2,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2005 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@cse.unsw.edu.au>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -30,7 +30,7 @@
 
 #include "mdadm.h"
 
-char Version[] = Name " - v2.2 - 5 December 2005\n";
+char Version[] = Name " - v2.3 - 2 February 2006\n";
 
 /*
  * File: ReadMe.c
@@ -182,15 +182,27 @@ char Usage[] =
 ;
 
 char Help[] =
+"mdadm is used for building, managing, and monitoring\n"
+"Linux md devices (aka RAID arrays)\n"
 "Usage: mdadm --create device options...\n"
+"            Create a new array from unused devices.\n"
 "       mdadm --assemble device options...\n"
+"            Assemble a previously created array.\n"
 "       mdadm --build device options...\n"
+"            Create or assemble an array without metadata.\n"
 "       mdadm --manage device options...\n"
+"            make changes to an existing array.\n"
 "       mdadm --misc options... devices\n"
+"            report on or modify various md related devices.\n"
 "       mdadm --monitor options...\n"
+"            Monitor one or more array for significant changes.\n"
 "       mdadm device options...\n"
-" mdadm is used for building, managing, and monitoring\n"
-" Linux md devices (aka RAID arrays)\n"
+"            Shorthand for --manage.\n"
+"Any parameter that does not start with '-' is treated as a device name\n"
+"or, for --examine-bitmap, a file name.\n"
+"The first such name is often the name of an md device.  Subsequent\n"
+"names are often names of component devices.\n"
+"\n"
 " For detailed help on the above major modes use --help after the mode\n"
 " e.g.\n"
 "         mdadm --assemble --help\n"
@@ -202,7 +214,7 @@ char OptionHelp[] =
 "Any parameter that does not start with '-' is treated as a device name\n"
 "or, for --examine-bitmap, a file name.\n"
 "The first such name is often the name of an md device.  Subsequent\n"
-"names are often names of component devices."
+"names are often names of component devices.\n"
 "\n"
 "Some common options are:\n"
 "  --help        -h   : General help message or, after above option,\n"
@@ -215,7 +227,7 @@ char OptionHelp[] =
 "  --force       -f   : Override normal checks and be more forceful\n"
 "\n"
 "  --assemble    -A   : Assemble an array\n"
-"  --build       -B   : Build a legacy array\n"
+"  --build       -B   : Build an array without metadata\n"
 "  --create      -C   : Create a new array\n"
 "  --detail      -D   : Display details of an array\n"
 "  --examine     -E   : Examine superblock on an array component\n"
