@@ -105,12 +105,51 @@ static void examine_super0(void *sbv)
 	printf("  Creation Time : %.24s\n", ctime(&atime));
 	c=map_num(pers, sb->level);
 	printf("     Raid Level : %s\n", c?c:"-unknown-");
-	if ((int)sb->level >= 0)
+	if ((int)sb->level >= 0) {
+		int ddsks=0;
 		printf("    Device Size : %d%s\n", sb->size, human_size((long long)sb->size<<10));
+		switch(sb->level) {
+		case 1: ddsks=1;break;
+		case 4:
+		case 5: ddsks = sb->raid_disks-1; break;
+		case 6: ddsks = sb->raid_disks-2; break;
+		case 10: ddsks = sb->raid_disks / (sb->layout&255) / ((sb->layout>>8)&255);
+		}
+		if (ddsks)
+			printf("     Array Size : %llu%s\n", (unsigned long long)ddsks * sb->size,
+			       human_size(ddsks*(long long)sb->size<<10));
+	}
 	printf("   Raid Devices : %d\n", sb->raid_disks);
 	printf("  Total Devices : %d\n", sb->nr_disks);
 	printf("Preferred Minor : %d\n", sb->md_minor);
 	printf("\n");
+	if (sb->minor_version > 90 && (sb->reshape_position+1) != 0) {
+		printf("  Reshape pos'n : %llu%s\n", sb->reshape_position/2, human_size((long long)sb->reshape_position<<9));
+		if (sb->delta_disks) {
+			printf("  Delta Devices : %d", sb->delta_disks);
+			if (sb->delta_disks)
+				printf(" (%d->%d)\n", sb->raid_disks-sb->delta_disks, sb->raid_disks);
+			else
+				printf(" (%d->%d)\n", sb->raid_disks, sb->raid_disks+sb->delta_disks);
+		}
+		if (sb->new_level != sb->level) {
+			c = map_num(pers, sb->new_level);
+			printf("      New Level : %s\n", c?c:"-unknown-");
+		}
+		if (sb->new_layout != sb->layout) {
+			if (sb->level == 5) {
+				c = map_num(r5layout, sb->new_layout);
+				printf("     New Layout : %s\n", c?c:"-unknown-");
+			}
+			if (sb->level == 10) {
+				printf("     New Layout : near=%d, far=%d\n",
+				       sb->new_layout&255, (sb->new_layout>>8)&255);
+			}
+		}
+		if (sb->new_chunk != sb->chunk_size)
+			printf("  New Chunksize : %d\n", sb->new_chunk);
+		printf("\n");
+	}
 	atime = sb->utime;
 	printf("    Update Time : %.24s\n", ctime(&atime));
 	printf("          State : %s\n",
@@ -359,6 +398,12 @@ static int update_super0(struct mdinfo *info, void *sbv, char *update, char *dev
 		sb->state &= ~(1<<MD_SB_CLEAN);
 		sb->recovery_cp = 0;
 	}
+	if (strcmp(update, "uuid") == 0) {
+		sb->set_uuid0 = info->uuid[0];
+		sb->set_uuid1 = info->uuid[1];
+		sb->set_uuid2 = info->uuid[2];
+		sb->set_uuid3 = info->uuid[3];
+	}
 
 	sb->sb_csum = calc_sb0_csum(sb);
 	return rv;
@@ -372,7 +417,7 @@ static __u64 event_super0(void *sbv)
 
 
 
-static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info, char *ignored_name)
+static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *ignored_name)
 {
 	mdp_super_t *sb = malloc(MD_SB_BYTES + sizeof(bitmap_super_t));
 	int spares;
@@ -401,6 +446,8 @@ static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info,
 		sb->set_uuid0 = random();
 	sb->ctime = time(0);
 	sb->level = info->level;
+	if (size != info->size)
+		return 0;
 	sb->size = info->size;
 	sb->nr_disks = info->nr_disks;
 	sb->raid_disks = info->raid_disks;
@@ -460,7 +507,7 @@ static int store_super0(struct supertype *st, int fd, void *sbv)
 			dsize = ((unsigned long long)size)<<9;
 	}
 
-	if (dsize < MD_RESERVED_SECTORS*2)
+	if (dsize < MD_RESERVED_SECTORS*2*512)
 		return 2;
 	
 	offset = MD_NEW_SIZE_SECTORS(dsize>>9);
@@ -657,7 +704,7 @@ static __u64 avail_size0(struct supertype *st, __u64 devsize)
 	return MD_NEW_SIZE_SECTORS(devsize);
 }
 
-static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int delay, int write_behind, int *sizep, int may_change)
+static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int delay, int write_behind, unsigned long long size, int may_change, int major)
 {
 	/*
 	 * The bitmap comes immediately after the superblock and must be 60K in size
@@ -665,7 +712,6 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 	 *
 	 * size is in K,  chunk is in bytes !!!
 	 */
-	unsigned long long size = *sizep;
 	unsigned long long bits;
 	unsigned long long max_bits = 60*1024*8;
 	unsigned long long min_chunk;
@@ -674,7 +720,7 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 
 	
 	min_chunk = 4096; /* sub-page chunks don't work yet.. */
-	bits = (size * 1024)/ min_chunk +1;
+	bits = (size * 512)/ min_chunk +1;
 	while (bits > max_bits) {
 		min_chunk *= 2;
 		bits = (bits+1)/2;
@@ -686,13 +732,13 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 
 	sb->state |= (1<<MD_SB_BITMAP_PRESENT);
 
-	memset(bms, sizeof(*bms), 0);
+	memset(bms, 0, sizeof(*bms));
 	bms->magic = __cpu_to_le32(BITMAP_MAGIC);
-	bms->version = __cpu_to_le32(BITMAP_MAJOR);
+	bms->version = __cpu_to_le32(major);
 	uuid_from_super0((int*)bms->uuid, sb);
 	bms->chunksize = __cpu_to_le32(chunk);
 	bms->daemon_sleep = __cpu_to_le32(delay);
-	bms->sync_size = __cpu_to_le64(size<<1);
+	bms->sync_size = __cpu_to_le64(size);
 	bms->write_behind = __cpu_to_le32(write_behind);
 
 
