@@ -58,7 +58,14 @@ struct mdp_superblock_1 {
 				 * NOTE: signed, so bitmap can be before superblock
 				 * only meaningful of feature_map[0] is set.
 				 */
-	__u8	pad1[128-100];	/* set to 0 when written */
+
+	/* These are only valid with feature bit '4' */
+	__u64	reshape_position;	/* next address in array-space for reshape */
+	__u32	new_level;	/* new level we are reshaping to		*/
+	__u32	delta_disks;	/* change in number of raid_disks		*/
+	__u32	new_layout;	/* new layout					*/
+	__u32	new_chunk;	/* new chunk size (bytes)			*/
+	__u8	pad1[128-124];	/* set to 0 when written */
 
 	/* constant this-device information - 64 bytes */
 	__u64	data_offset;	/* sector start of data, often 0 */
@@ -89,6 +96,15 @@ struct mdp_superblock_1 {
 	__u16	dev_roles[0];	/* role in array, or 0xffff for a spare, or 0xfffe for faulty */
 };
 
+/* feature_map bits */
+#define MD_FEATURE_BITMAP_OFFSET	1
+#define	MD_FEATURE_RECOVERY_OFFSET	2 /* recovery_offset is present and
+					   * must be honoured
+					   */
+#define	MD_FEATURE_RESHAPE_ACTIVE	4
+
+#define	MD_FEATURE_ALL			(1|2|4)
+
 #ifndef offsetof
 #define offsetof(t,f) ((int)&(((t*)0)->f))
 #endif
@@ -132,7 +148,8 @@ static void examine_super1(void *sbv)
 	char *c;
 
 	printf("          Magic : %08x\n", __le32_to_cpu(sb->magic));
-	printf("        Version : %02d.%02d\n", 1, __le32_to_cpu(sb->feature_map));
+	printf("        Version : %02d\n", 1);
+	printf("    Feature Map : 0x%x\n", __le32_to_cpu(sb->feature_map));
 	printf("     Array UUID : ");
 	for (i=0; i<16; i++) {
 		if ((i&3)==0 && i != 0) printf(":");
@@ -148,10 +165,26 @@ static void examine_super1(void *sbv)
 	printf("   Raid Devices : %d\n", __le32_to_cpu(sb->raid_disks));
 	printf("\n");
 	printf("    Device Size : %llu%s\n", (unsigned long long)sb->data_size, human_size(sb->data_size<<9));
+	if (__le32_to_cpu(sb->level) >= 0) {
+		int ddsks=0;
+		switch(__le32_to_cpu(sb->level)) {
+		case 1: ddsks=1;break;
+		case 4:
+		case 5: ddsks = sb->raid_disks-1; break;
+		case 6: ddsks = sb->raid_disks-2; break;
+		case 10: ddsks = sb->raid_disks / (sb->layout&255) / ((sb->layout>>8)&255);
+		}
+		if (ddsks)
+			printf("     Array Size : %llu%s\n", ddsks*(unsigned long long)sb->size, human_size(ddsks*sb->size<<9));
+		if (sb->size != sb->data_size)
+			printf("      Used Size : %llu%s\n", (unsigned long long)sb->size, human_size(sb->size<<9));
+	}
 	if (sb->data_offset)
 		printf("    Data Offset : %llu sectors\n", (unsigned long long)__le64_to_cpu(sb->data_offset));
 	if (sb->super_offset)
 		printf("   Super Offset : %llu sectors\n", (unsigned long long)__le64_to_cpu(sb->super_offset));
+	if (__le32_to_cpu(sb->feature_map) & MD_FEATURE_RECOVERY_OFFSET)
+		printf("Recovery Offset : %llu sectors\n", (unsigned long long)__le64_to_cpu(sb->recovery_offset));
 	printf("          State : %s\n", (__le64_to_cpu(sb->resync_offset)+1)? "active":"clean");
 	printf("    Device UUID : ");
 	for (i=0; i<16; i++) {
@@ -159,6 +192,43 @@ static void examine_super1(void *sbv)
 		printf("%02x", sb->device_uuid[i]);
 	}
 	printf("\n");
+	printf("\n");
+	if (sb->feature_map & __cpu_to_le32(MD_FEATURE_BITMAP_OFFSET)) {
+		printf("Internal Bitmap : %ld sectors from superblock\n",
+		       (long)__le32_to_cpu(sb->bitmap_offset));
+	}
+	if (sb->feature_map & __le32_to_cpu(MD_FEATURE_RESHAPE_ACTIVE)) {
+		printf("  Reshape pos'n : %llu%s\n", __le64_to_cpu(sb->reshape_position)/2,
+		       human_size(__le64_to_cpu(sb->reshape_position)<<9));
+		if (__le32_to_cpu(sb->delta_disks)) {
+			printf("  Delta Devices : %d", __le32_to_cpu(sb->delta_disks));
+			if (__le32_to_cpu(sb->delta_disks))
+				printf(" (%d->%d)\n",
+				       __le32_to_cpu(sb->raid_disks)-__le32_to_cpu(sb->delta_disks),
+				       __le32_to_cpu(sb->raid_disks));
+			else
+				printf(" (%d->%d)\n", __le32_to_cpu(sb->raid_disks),
+				       __le32_to_cpu(sb->raid_disks)+__le32_to_cpu(sb->delta_disks));
+		}
+		if (__le32_to_cpu(sb->new_level) != __le32_to_cpu(sb->level)) {
+			c = map_num(pers, __le32_to_cpu(sb->new_level));
+			printf("      New Level : %s\n", c?c:"-unknown-");
+		}
+		if (__le32_to_cpu(sb->new_layout) != __le32_to_cpu(sb->layout)) {
+			if (__le32_to_cpu(sb->level) == 5) {
+				c = map_num(r5layout, __le32_to_cpu(sb->new_layout));
+				printf("     New Layout : %s\n", c?c:"-unknown-");
+			}
+			if (__le32_to_cpu(sb->level) == 10) {
+				printf("     New Layout : near=%d, far=%d\n",
+				       __le32_to_cpu(sb->new_layout)&255,
+				       (__le32_to_cpu(sb->new_layout)>>8)&255);
+			}
+		}
+		if (__le32_to_cpu(sb->new_chunk) != __le32_to_cpu(sb->chunksize))
+			printf("  New Chunksize : %dK\n", __le32_to_cpu(sb->new_chunk)/2);
+		printf("\n");
+	}
 	if (sb->devflags) {
 		printf("      Flags :");
 		if (sb->devflags & WriteMostly1)
@@ -189,10 +259,12 @@ static void examine_super1(void *sbv)
 	case 0:
 	case 4:
 	case 5:
-		printf("     Chunk Size : %dK\n", __le32_to_cpu(sb->chunksize/2));
+	case 6:
+	case 10:
+		printf("     Chunk Size : %dK\n", __le32_to_cpu(sb->chunksize)/2);
 		break;
 	case -1:
-		printf("       Rounding : %dK\n", __le32_to_cpu(sb->chunksize/2));
+		printf("       Rounding : %dK\n", __le32_to_cpu(sb->chunksize)/2);
 		break;
 	default: break;
 	}
@@ -382,7 +454,7 @@ static int update_super1(struct mdinfo *info, void *sbv, char *update, char *dev
 		sb->resync_offset = ~0ULL;
 	}
 	if (strcmp(update, "uuid") == 0)
-		memcmp(sb->set_uuid, info->uuid, 16);
+		memcpy(sb->set_uuid, info->uuid, 16);
 
 	sb->sb_csum = calc_sb_1_csum(sb);
 	return rv;
@@ -395,7 +467,7 @@ static __u64 event_super1(void *sbv)
 	return __le64_to_cpu(sb->events);
 }
 
-static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info, char *name)
+static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *name)
 {
 	struct mdp_superblock_1 *sb = malloc(1024 + sizeof(bitmap_super_t));
 	int spares;
@@ -434,7 +506,7 @@ static int init_super1(struct supertype *st, void **sbp, mdu_array_info_t *info,
 	sb->ctime = __cpu_to_le64((unsigned long long)time(0));
 	sb->level = __cpu_to_le32(info->level);
 	sb->layout = __cpu_to_le32(info->layout);
-	sb->size = __cpu_to_le64(info->size*2ULL);
+	sb->size = __cpu_to_le64(size*2ULL);
 	sb->chunksize = __cpu_to_le32(info->chunk_size>>9);
 	sb->raid_disks = __cpu_to_le32(info->raid_disks);
 
@@ -477,13 +549,21 @@ static int store_super1(struct supertype *st, int fd, void *sbv)
 	struct mdp_superblock_1 *sb = sbv;
 	unsigned long long sb_offset;
 	int sbsize;
-	long size;
+	unsigned long size;
+	unsigned long long dsize;
 
-	if (ioctl(fd, BLKGETSIZE, &size))
-		return 1;
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = (unsigned long long)size;
+	} else
+		dsize >>= 9;
 
-
-	if (size < 24)
+	if (dsize < 24)
 		return 2;
 
 	/*
@@ -496,7 +576,7 @@ static int store_super1(struct supertype *st, int fd, void *sbv)
 	 */
 	switch(st->minor_version) {
 	case 0:
-		sb_offset = size;
+		sb_offset = dsize;
 		sb_offset -= 8*2;
 		sb_offset &= ~(4*2-1);
 		break;
@@ -542,7 +622,8 @@ static int write_init_super1(struct supertype *st, void *sbv,
 	int rfd;
 	int rv;
 
-	long size;
+	unsigned long size, space;
+	unsigned long long dsize, array_size;
 	long long sb_offset;
 
 
@@ -578,12 +659,18 @@ static int write_init_super1(struct supertype *st, void *sbv,
 		free(refsb);
 	}
 
-	if (ioctl(fd, BLKGETSIZE, &size)) {
-		close(fd);
-		return 1;
-	}
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size))
+			return 1;
+		else
+			dsize = size;
+	} else
+		dsize >>= 9;
 
-	if (size < 24) {
+	if (dsize < 24) {
 		close(fd);
 		return 2;
 	}
@@ -596,26 +683,40 @@ static int write_init_super1(struct supertype *st, void *sbv,
 	 * 0: At least 8K, but less than 12K, from end of device
 	 * 1: At start of device
 	 * 2: 4K from start of device.
+	 * Depending on the array size, we might leave extra space
+	 * for a bitmap.
 	 */
+	array_size = __le64_to_cpu(sb->size);
 	switch(st->minor_version) {
 	case 0:
-		sb_offset = size;
+		sb_offset = dsize;
 		sb_offset -= 8*2;
 		sb_offset &= ~(4*2-1);
 		sb->super_offset = __cpu_to_le64(sb_offset);
 		sb->data_offset = __cpu_to_le64(0);
-		sb->data_size = __cpu_to_le64(sb_offset);
+		if (sb_offset-64*2 >= array_size && array_size > 8*1024*1024*2)
+			sb->data_size = __cpu_to_le64(sb_offset-64*2);
+		else
+			sb->data_size = __cpu_to_le64(sb_offset);
 		break;
 	case 1:
 		sb->super_offset = __cpu_to_le64(0);
-		sb->data_offset = __cpu_to_le64(4*2); /* leave 4k for super and bitmap */
-		sb->data_size = __cpu_to_le64(size - 4*2);
+		if (dsize - 64*2 >= array_size && array_size > 8*1024*1024*2)
+			space = 64*2;
+		else
+			space = 4*2;
+		sb->data_offset = __cpu_to_le64(space); /* leave space for super and bitmap */
+		sb->data_size = __cpu_to_le64(dsize - space);
 		break;
 	case 2:
 		sb_offset = 4*2;
+		if (dsize - 4*2 - 64*2 >= array_size && array_size > 8*1024*1024*2)
+			space = 64*2;
+		else
+			space = 4*2;
 		sb->super_offset = __cpu_to_le64(sb_offset);
-		sb->data_offset = __cpu_to_le64(sb_offset+4*2);
-		sb->data_size = __cpu_to_le64(size - 4*2 - 4*2);
+		sb->data_offset = __cpu_to_le64(sb_offset+space);
+		sb->data_size = __cpu_to_le64(dsize - 4*2 - space);
 		break;
 	default:
 		return -EINVAL;
@@ -672,6 +773,7 @@ static int compare_super1(void **firstp, void *secondv)
 static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 {
 	unsigned long size;
+	unsigned long long dsize;
 	unsigned long long sb_offset;
 	struct mdp_superblock_1 *super;
 
@@ -709,17 +811,24 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 		st->ss = NULL;
 		return 2;
 	}
-	if (ioctl(fd, BLKGETSIZE, &size)) {
-		if (devname)
-			fprintf(stderr, Name ": cannot find device size for %s: %s\n",
-				devname, strerror(errno));
-		return 1;
-	}
+#ifdef BLKGETSIZE64
+	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
+#endif
+	{
+		if (ioctl(fd, BLKGETSIZE, &size)) {
+			if (devname)
+				fprintf(stderr, Name ": cannot find device size for %s: %s\n",
+					devname, strerror(errno));
+			return 1;
+		}
+		dsize = size;
+	} else
+		dsize >>= 9;
 
-	if (size < 24) {
+	if (dsize < 24) {
 		if (devname)
-			fprintf(stderr, Name ": %s is too small for md: size is %lu sectors.\n",
-				devname, size);
+			fprintf(stderr, Name ": %s is too small for md: size is %llu sectors.\n",
+				devname, dsize);
 		return 1;
 	}
 
@@ -733,7 +842,7 @@ static int load_super1(struct supertype *st, int fd, void **sbp, char *devname)
 	 */
 	switch(st->minor_version) {
 	case 0:
-		sb_offset = size;
+		sb_offset = dsize;
 		sb_offset -= 8*2;
 		sb_offset &= ~(4*2-1);
 		break;
@@ -802,7 +911,8 @@ static struct supertype *match_metadata_desc1(char *arg)
 	st->ss = &super1;
 	st->max_devs = 384;
 	if (strcmp(arg, "1") == 0 ||
-	    strcmp(arg, "1.0") == 0) {
+	    strcmp(arg, "1.0") == 0 ||
+	    strcmp(arg, "default/large") == 0) {
 		st->minor_version = 0;
 		return st;
 	}
@@ -828,6 +938,14 @@ static __u64 avail_size1(struct supertype *st, __u64 devsize)
 	if (devsize < 24)
 		return 0;
 
+	/* if the device is bigger than 8Gig, save 64k for bitmap usage,
+	 * if biffer than 200Gig, save 128k
+	 */
+	if (devsize > 200*1024*1024*2)
+		devsize -= 128*2;
+	else if (devsize > 8*1024*1024*2)
+		devsize -= 64*2;
+
 	switch(st->minor_version) {
 	case 0:
 		/* at end */
@@ -844,7 +962,8 @@ static __u64 avail_size1(struct supertype *st, __u64 devsize)
 
 static int
 add_internal_bitmap1(struct supertype *st, void *sbv,
-		     int chunk, int delay, int write_behind, unsigned long long size, int may_change, int major)
+		     int chunk, int delay, int write_behind, unsigned long long size,
+		     int may_change, int major)
 {
 	/*
 	 * If not may_change, then this is a 'Grow', and the bitmap
@@ -878,11 +997,13 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 		chunk = min_chunk;
 	else if (chunk < min_chunk)
 		return 0; /* chunk size too small */
+	if (chunk == 0) /* rounding problem */
+		return 0;
 
 	sb->bitmap_offset = __cpu_to_le32(2);
 
 	sb->feature_map = __cpu_to_le32(__le32_to_cpu(sb->feature_map) | 1);
-	memset(bms, sizeof(*bms), 0);
+	memset(bms, 0, sizeof(*bms));
 	bms->magic = __cpu_to_le32(BITMAP_MAGIC);
 	bms->version = __cpu_to_le32(major);
 	uuid_from_super1((int*)bms->uuid, sb);
@@ -898,14 +1019,14 @@ add_internal_bitmap1(struct supertype *st, void *sbv,
 void locate_bitmap1(struct supertype *st, int fd, void *sbv)
 {
 	unsigned long long offset;
-	struct mdp_superblock_1 *sb;
+	struct mdp_superblock_1 *sb = NULL;
 
-	if (!sbv)
-		if (st->ss->load_super(st, fd, sbv, NULL))
+	if (sbv)
+		sb = sbv;
+	else
+		if (st->ss->load_super(st, fd, (void **)&sb, NULL))
 			return; /* no error I hope... */
 
-	sb = sbv;
-
 	offset = __le64_to_cpu(sb->super_offset);
 	offset += (long) __le32_to_cpu(sb->bitmap_offset);
 	if (!sbv)
